#version 420 core //otherwise layout(binding will not work

in VS_OUT {
    vec3 FragPos;
    vec2 TexCoords;
    vec4 TangentLightPos1; //cannot pass array -> error
	vec4 TangentLightPos2;
	vec4 TangentLightPos3;
	vec4 TangentLightPos4;
    vec3 TangentViewPos;
    vec3 TangentFragPos;
} fs_in;

vec4 TangentLightPos[4];

//textures
uniform sampler2D material_texture;
uniform sampler2D height_texture;
uniform sampler2D normal_texture;

//scale for size of parallax mapping effect
uniform float parallaxScale; //height_scale

uniform bool enableParallax;
uniform bool enableShadow;

uniform float ambientIntensity;
uniform float diffuseIntensity;
uniform float specularIntensity;

uniform vec3	matSpecularReflectanceColor;	// = vec3(1, 1, 1);
uniform float	matShininess;				// = 64;

out vec4 outputColor;



/////////////////////
// Implements Parallax Mapping technique
// Returns modified texture coordinates, and last used depth
vec2 ParallaxMapping(vec2 texCoords, vec3 viewDir)
{
	// number of depth layers
   	const float minLayers = 5;
	const float maxLayers = 15;
	float numLayers = mix(maxLayers, minLayers, abs(dot(vec3(0, 0, 1), viewDir)));

    // calculate the size of each layer
    float layerDepth = 1.0 / numLayers;
    // depth of current layer
    float currentLayerDepth = 0.0;
    // the amount to shift the texture coordinates per layer (from vector P)
    vec2 P = (viewDir.xy / viewDir.z) * parallaxScale; 
    vec2 deltaTexCoords = P / numLayers;

	// get initial values
	vec2  currentTexCoords     = texCoords;
	float currentDepthMapValue = texture(height_texture, currentTexCoords).r;
  
	while(currentLayerDepth < currentDepthMapValue)
	{
		// shift texture coordinates along direction of P
		currentTexCoords -= deltaTexCoords;
		// get depthmap value at current texture coordinates
		currentDepthMapValue = texture(height_texture, currentTexCoords).r;  
		// get depth of next layer
		currentLayerDepth += layerDepth;  
	}

	return currentTexCoords;
}

/////////////////////
// Calculates soft shadows
//
float parallaxSoftShadowMultiplier(in vec3 lightDir[4], in vec2 initialTexCoord)
{
	float shadowMultiplier[4] = {1,1,1,1};
	// number of depth layers
	const float minLayers = 15;
	const float maxLayers = 30;

	// calculate lighting only for surface oriented to the light source
	for(int i = 0; i < 4; i++)
	{
		if(dot(vec3(0, 0, 1), lightDir[i]) > 0)
		{
			// calculate initial parameters
			float numSamplesUnderSurface	= 0;
			shadowMultiplier[i]	= 0;
			float numLayers = mix(maxLayers, minLayers, abs(dot(vec3(0, 0, 1), lightDir[i])));
			// calculate the size of each layer
			float layerDepth = 1.0 / numLayers;
			// the amount to shift the texture coordinates per layer (from vector P)
			vec2 P = (lightDir[i].xy / lightDir[i].z) * parallaxScale; 
			vec2 deltaTexCoords = P / numLayers;
			float initialDepth = texture(height_texture, initialTexCoord).r;

			// current parameters
			float currentLayerDepth	= initialDepth - layerDepth;
			vec2 currentTextureCoords	= initialTexCoord + deltaTexCoords;
			float depthFromTexture	= texture(height_texture, currentTextureCoords).r;
			int stepIndex	= 1;

			// while point is below depth 0.0 )
			while(currentLayerDepth > 0)
			{
				// if point is under the surface
				if(depthFromTexture < currentLayerDepth)
				{
					// calculate partial shadowing factor
					numSamplesUnderSurface	+= 1;
					float newShadowMultiplier	= (currentLayerDepth - depthFromTexture) *
														(1.0 - stepIndex / numLayers);
					shadowMultiplier[i]	= max(shadowMultiplier[i], newShadowMultiplier);
				}

				// offset to the next layer
				stepIndex += 1;
				currentLayerDepth -= layerDepth;
				currentTextureCoords += deltaTexCoords;
				depthFromTexture = texture(height_texture, currentTextureCoords).r;
			}

			// Shadowing factor should be 1 if there were no points under the surface
			if(numSamplesUnderSurface < 1)
			{
				shadowMultiplier[i] = 1;
			}
			else
			{
				shadowMultiplier[i] = 1.0 - shadowMultiplier[i];
			}
		}
	}
	
	float shadowMultiplierSum = 0.0f;
	//combine shadwoMultipliers
	for(int i = 0; i < 4; i++)
	{
		float distance = length(TangentLightPos[i].xyz - fs_in.TangentFragPos);
		float lightPower = TangentLightPos[i].w;
		float attenuation = 1.0 / distance;
		//attenuation *= lightPower;
		shadowMultiplier[i] *= attenuation;
		shadowMultiplierSum += shadowMultiplier[i];
	}

	return shadowMultiplierSum;
}

/////////////////////
// Calculates lighting by Blinn-Phong model and Normal Mapping
// Returns color of the fragment
vec4 normalMappingLighting(in vec2 texCoord, in vec3 lightDir[4], in vec3 viewDir, float shadowMultiplier)
{
	// restore normal from normal map
   vec3 normalDir = normalize(texture(normal_texture, texCoord).xyz * 2 - 1);
   vec3 diffuseColor = texture(material_texture, texCoord).rgb;
   vec3 ambientColor = diffuseColor;
   vec3 specularColor = matSpecularReflectanceColor;
   vec4 resColor;
   
   vec3 ambient = ambientColor * ambientIntensity;

   for(int i = 0; i < 4; i++)
   {
		//Diffuse
	   float diff = max(dot(lightDir[i], normalDir), 0.0);
	   vec3 diffuse = diffuseColor * diff * diffuseIntensity;
   
	   //specular lighting
	   float spec = 0;
	   
	   if(dot(normalDir,lightDir[i]) > 0.2)
	   {
			vec3 reflectDir = reflect(-lightDir[i], normalDir);
			spec = pow(max(dot(reflectDir,viewDir), 0.0), matShininess);
	   }
	   vec3 specular = specularColor * spec * specularIntensity;
	   
	   //attenuation
	   float distance = length(TangentLightPos[i].xyz - fs_in.TangentFragPos);
	   float attenuation = 1.0 / (distance * distance);

	   float lightPower = TangentLightPos[i].w;

		resColor.rgb += ambient +
						(diffuse + specular) * attenuation * lightPower;
   }
   resColor.a = 1;
   
   //DEBUG
   //vec4 resColor;
   //resColor.rgb = D;
   //resColor.a = 1;

   return resColor;
}

/////////////////////
// Entry point for Parallax Mapping shader
void main()
{
	// Offset texture coordinates with Parallax Mapping
    vec3 viewDir   = normalize(fs_in.TangentViewPos - fs_in.TangentFragPos);
	
	TangentLightPos[0] = fs_in.TangentLightPos1;
	TangentLightPos[1] = fs_in.TangentLightPos2;
	TangentLightPos[2] = fs_in.TangentLightPos3;
	TangentLightPos[3] = fs_in.TangentLightPos4;

	vec3 lightDir[4];
	for(int i = 0; i < 4; i++)
	{
		lightDir[i]  = normalize(TangentLightPos[i].xyz - fs_in.TangentFragPos);
	}

	//vec3[4] lightDir;
	//lightDir[0] = normalize(fs_in.WorldLightPos - fs_in.FragPos);
    

	vec2 texCoords = fs_in.TexCoords;
	if(enableParallax == true) {
		texCoords = ParallaxMapping(fs_in.TexCoords,  viewDir);
	}

    // then sample textures with new texture coords
	//vec3 diffuse = texture(material_texture, texCoords).rgb;
	//vec4 normal  = texture(normal_texture, texCoords);
	//normal = normalize(normal * 2.0 - 1.0);

	//calculate softShadows
	float shadowMultiplier = 1;
	if(enableShadow == true) {
		shadowMultiplier = parallaxSoftShadowMultiplier(lightDir, fs_in.TexCoords);
	}

	//calculate lighting
	outputColor = normalMappingLighting(texCoords, lightDir, viewDir, shadowMultiplier);
}